<?php

namespace Drupal\block_usage_report\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityListBuilderInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns output for the block usage report.
 */
class BlockUsageReportController extends ControllerBase {

  /**
   * The block storage provider.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $blockStorage;

  /**
   * The block content storage provider.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $blockContentStorage;

  /**
   * The block list builder.
   *
   * @var \Drupal\Core\Entity\EntityListBuilderInterface
   */
  protected EntityListBuilderInterface $blockListBuilder;

  /**
   * The block content list builder.
   *
   * @var \Drupal\Core\Entity\EntityListBuilderInterface
   */
  protected EntityListBuilderInterface $blockContentListBuilder;

  /**
   * The fixed block content storage provider, if the associated module exists.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface|null
   */
  protected ?EntityStorageInterface $fixedBlockContentStorage;

  /**
   * Constructor for this controller.
   *
   * @param \Drupal\Core\Entity\EntityStorageInterface $block_storage
   *   The block storage provider.
   * @param \Drupal\Core\Entity\EntityStorageInterface $block_content_storage
   *   The block content storage provider.
   * @param \Drupal\Core\Entity\EntityListBuilderInterface $block_list_builder
   *   The block list builder.
   * @param \Drupal\Core\Entity\EntityListBuilderInterface $block_content_list_builder
   *   The block content list builder.
   * @param \Drupal\Core\Entity\EntityStorageInterface|null $fixed_block_content_storage
   *   The fixed block content storage provider, or NULL if the associated
   *   module does not exist.
   */
  public function __construct(
    EntityStorageInterface $block_storage,
    EntityStorageInterface $block_content_storage,
    EntityListBuilderInterface $block_list_builder,
    EntityListBuilderInterface $block_content_list_builder,
    ?EntityStorageInterface $fixed_block_content_storage
  ) {
    $this->blockStorage = $block_storage;
    $this->blockContentStorage = $block_content_storage;
    $this->blockListBuilder = $block_list_builder;
    $this->blockContentListBuilder = $block_content_list_builder;
    $this->fixedBlockContentStorage = $fixed_block_content_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\Core\Extension\ModuleHandlerInterface $module_handler */
    $module_handler = $container->get('module_handler');
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_mgr */
    $entity_type_mgr = $container->get('entity_type.manager');

    $fixed_block_storage = NULL;
    if (
      $module_handler->moduleExists('fixed_block_content')
      && $entity_type_mgr->hasHandler('fixed_block_content', 'storage')
    ) {
      $fixed_block_storage = $entity_type_mgr->getStorage('fixed_block_content');
    }

    return new static(
      $entity_type_mgr->getStorage('block'),
      $entity_type_mgr->getStorage('block_content'),
      $entity_type_mgr->getListBuilder('block'),
      $entity_type_mgr->getListBuilder('block_content'),
      $fixed_block_storage
    );
  }

  /**
   * Finds a content entity referenced by a UUID, given a collection.
   *
   * @param string $uuid
   *   The UUID to search for.
   * @param \Drupal\Core\Entity\ContentEntityInterface[] $entities
   *   The collection of entities to search for.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface|null
   *   THe matching entity, or NULL if not found.
   */
  protected function getEntityByUuid(string $uuid, array $entities): ?ContentEntityInterface {
    foreach ($entities as $entity) {
      if ($entity->uuid() == $uuid) {
        return $entity;
      }
    }
    return NULL;
  }

  /**
   * Builds the output of the report.
   *
   * @return array
   *   The render-array for the report.
   */
  public function build(): array {
    /** @var \Drupal\Core\Entity\ContentEntityInterface[] $bc_blocks */
    $bc_blocks = $this->blockContentStorage->loadMultiple();
    $bc_uuids_found = [];

    /** @var \Drupal\block\Entity\Block[] $blocks */
    $blocks = $this->blockStorage->loadMultiple();
    $public_theme_name = $this->config('system.theme')->get('default');

    $header = [
      $this->t('Name'),
      $this->t('Provider'),
      $this->t('Type'),
      $this->t('Region'),
      $this->t('Operations'),
    ];

    $enabled_rows = $disabled_rows = [];

    foreach ($blocks as $block) {
      $block_theme = $block->getTheme();
      if ($block_theme != $public_theme_name) {
        continue;
      }
      $settings = $block->get('settings');
      $name = $block->label() . ' (' . $block->id() . ')';
      $provider = $settings['provider'];
      $region = $block->getRegion();
      $status = $block->status();
      $plugin_id = $block->getPluginId();
      if (str_contains($plugin_id, ':')) {
        $type = explode(':', $plugin_id)[1];
      }
      else {
        $type = '';
      }

      if ($provider == 'block_content') {
        $uuid = explode(':', $plugin_id)[1];
        $block_content = $this->getEntityByUuid($uuid, $bc_blocks);
        $type = $block_content ? $block_content->bundle() : $this->t('[broken/missing]');
        $bc_uuids_found[] = $uuid;
      }
      elseif (
        $provider == 'fixed_block_content'
        && !empty($type)
        && !empty($this->fixedBlockContentStorage)
        && $fixed_block = $this->fixedBlockContentStorage->load($type)
      ) {
        // Register the fixed block's associated custom block as in-use.
        /** @var \Drupal\fixed_block_content\Entity\FixedBlockContent $fixed_block */
        /** @var \Drupal\block_content\BlockContentInterface $block_content */
        $block_content = $fixed_block->getBlockContent();
        // Show the user the bundle and ID of the associated custom block.
        $type .= '-' . $block_content->bundle() . '-' . $block_content->id();
        $bc_uuids_found[] = $block_content->uuid();
      }

      $row = [
        $name,
        $provider,
        $type,
        $region,
        [
          'data' => [
            '#type' => 'operations',
            '#links' => $this->blockListBuilder->getOperations($block),
          ],
        ],
      ];

      if ($status) {
        $enabled_rows[] = $row;
      }
      else {
        $disabled_rows[] = $row;
      }

    }

    // Sort first by provider, then by type, then by name.
    $sorter = function ($a, $b) {
      if ($a[1] != $b[1]) {
        return $a[1] <=> $b[1];
      }
      if ($a[2] != $b[2]) {
        return $a[2] <=> $b[2];
      }
      return $a[0] <=> $b[0];
    };

    usort($enabled_rows, $sorter);
    usort($disabled_rows, $sorter);

    $unplaced_blocks = [];
    foreach ($bc_blocks as $bc_block) {
      if (!in_array($bc_block->uuid(), $bc_uuids_found)) {
        $unplaced_blocks[] = [
          $bc_block->label() . ' (' . $bc_block->id() . ')',
          $bc_block->bundle(),
          [
            'data' => [
              '#type' => 'operations',
              '#links' => $this->blockContentListBuilder->getOperations($bc_block),
            ],
          ],
        ];
      }
    }
    if (!empty($unplaced_blocks)) {
      // Sort first by type, then by name.
      usort($unplaced_blocks, function ($a, $b) {
        if ($a[1] != $b[1]) {
          return $a[1] <=> $b[1];
        }
        return $a[0] <=> $b[0];
      });
    }

    $out = [];
    if (!empty($enabled_rows)) {
      $out['enabled_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Enabled blocks'),
        '#open' => FALSE,
      ];
      $out['enabled_wrapper']['enabled'] = [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $enabled_rows,
      ];
    }
    if (!empty($disabled_rows)) {
      $out['disabled_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Disabled blocks'),
        '#open' => FALSE,
      ];
      $out['disabled_wrapper']['disabled'] = [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $disabled_rows,
      ];
    }
    if (!empty($unplaced_blocks)) {
      $out['unplaced_wrapper'] = [
        '#type' => 'details',
        '#title' => $this->t('Unplaced custom blocks'),
        '#open' => FALSE,
      ];
      $out['unplaced_wrapper']['unplaced'] = [
        '#theme' => 'table',
        '#header' => [
          $this->t('Name'),
          $this->t('Type'),
          $this->t('Operations'),
        ],
        '#rows' => $unplaced_blocks,
      ];
    }
    return $out;

  }

}
